from threading import Thread
import socket
import time

__all__ = ["OPCUAConnection"]

class OPCUAConnection(Thread):
    """
      Connects to OPC-UA in the foreground or background, and sends HELLO
      messages to keep a check on the connection. On connection failure, reconnects once.
    """

    def __init__(self, client, on_func, fault_func, streams, try_interval=2):
        super().__init__(daemon=True)

        self.client = client
        self.on_func = on_func
        self.fault_func = fault_func
        self.try_interval = try_interval
        self.streams = streams
        self.stopping = False
        self.connected = False

    def _servername(self):
        return self.client.server_url.geturl()

    def connect(self):
        try:
            self.streams.debug_stream("Connecting to server %s", self._servername())
            self.client.connect()
            self.connected = True
            self.streams.debug_stream("Connected to server. Initialising.")
            return True
        except socket.error as e:
            self.streams.error_stream("Could not connect to server %s: %s", self._servername(), e)
            return False

    def disconnect(self):
        self.connected = False # always force a reconnect, regardless of a successful disconnect

        try:
            self.client.disconnect()
        except Exception as e:
            self.streams.error_stream("Disconnect from OPC-UA server %s failed: %s", self._servername(), e)

    def run(self):
        while not self.stopping:
            # keep trying to connect
            if not self.connected:
                if self.connect():
                    self.on_func()
                else:
                    # we retry only once, to catch exotic network issues. if the infra or hardware is down,
                    # our device cannot help, and must be reinitialised after the infra or hardware is fixed.
                    self.fault_func()
                    return

            # keep checking if the connection is still alive
            try:
                while not self.stopping:
                    self.client.send_hello()
                    time.sleep(self.try_interval)
            except Exception as e:
                self.streams.error_stream("Lost connection to server %s: %s", self._servername(), e)

                # technically, we may not have dropped the connection, but encounter a different error. so explicitly disconnect.
                self.disconnect()

                # signal that we're disconnected
                self.fault_func()

    def stop(self):
        """
          Stop connecting & disconnect. Can take a few seconds for the timeouts to hit.
        """

        if not self.ident:
            # have not yet been started, so nothing to do
            return

        self.stopping = True
        self.join()

        self.disconnect()
